########################################################################
# Introscope EPAgent Process (Windows) Perl Library
#                                                                      
# Wily Technology (tm) Introscope(R) Version 4 Release 4.2
# Copyright (c) 1998 - 2003 Wily Technology, Inc. All Rights Reserved.
# Introscope(R) is a registered trademark of Wily Technology, Inc.
#
########################################################################
# Process (Windows) Module
########################################################################

package Wily::WinProcess;

use strict;

#----------------
# Import Modules
#----------------
use FindBin;
use lib ("$FindBin::Bin", "$FindBin::Bin/.");
use Wily::PrintMetric;
use Win32::PerfLib;
use Getopt::Long;
use Carp;

# Define Sub-routines
#
sub reportDeadProcess($\%\%) {
	my $resource = shift();
	my %process = %{ shift() };
	my %process_data = %{ shift() };
	### print "### reportDeadProcess: Application $process{UNIQUE_ID} no longer exists!\n";
	Wily::PrintMetric::printMetric( type        => $process_data{ALIVE}{datatype},
					resource    => $resource,
					subresource => $process{UNIQUE_ID},
					name        => $process_data{ALIVE}{description},
					value       => 0,
	);
	if ( defined $process{PID} ) {
		Wily::PrintMetric::printMetric( type        => $process_data{PID}{datatype},
						resource    => $resource,
						subresource => $process{UNIQUE_ID},
						name        => $process_data{PID}{description},
						value       => $process{PID},
		);
	}
	# reset some numerical values
	while ((my $key,my $value) = each %process) {
		if ( defined $process_data{ $key } && $process_data{ $key }{datatype} =~ /Counter/i ) { 
			Wily::PrintMetric::printMetric( type        => $process_data{ $key }{datatype},
							resource    => $resource,
							subresource => $process{UNIQUE_ID},
							name        => $process_data{ $key }{description},
							value       => 0,
			);
		}
	}
}

sub reportAliveProcess($\%\%) {
	my $resource = shift();
	my %process = %{ shift() };
	my %process_data = %{ shift() };
	### print "### reportAliveProcess: Application $process{UNIQUE_ID} is alive!\n";
	Wily::PrintMetric::printMetric( type        => $process_data{ALIVE}{datatype},
					resource    => $resource,
					subresource => $process{UNIQUE_ID},
					name        => $process_data{ALIVE}{description},
					value       => 1,
	);
	while ((my $key,my $value) = each %process) {
		if ( defined $value && defined $process_data{ $key } ) {
			### print "### reportAliveProcess: $key=$value\n";
			Wily::PrintMetric::printMetric( type        => $process_data{ $key }{datatype},
							resource    => $resource,
							subresource => $process{UNIQUE_ID},
							name        => $process_data{ $key }{description},
							value       => $value,
			);
		}
	}
}

sub pingProcesses(\%$) { 
	my %process_data = %{ shift() };
	my $compareTimeDiff = shift();
	if ( not defined $compareTimeDiff || $compareTimeDiff <= 0 ) {
		$compareTimeDiff = 5;
	}
	my $winPerfLib = new Win32::PerfLib( ); # no servername defined (i.e. localhost )
	my $winProcRef0 = {};
	my $winProcRef1 = {};
	$winPerfLib->GetObjectList(230, $winProcRef0);
	# Sleep the specified amount
	# (must occur between the two calls to Windows perflib
	#
	sleep $compareTimeDiff;
	$winPerfLib->GetObjectList(230, $winProcRef1);
	my $winInstanceRef0 = $winProcRef0->{Objects}->{230}->{Instances};
	my $winInstanceRef1 = $winProcRef1->{Objects}->{230}->{Instances};
	$winPerfLib->Close();
	my %processes;
	my $foundPID = 0;
	foreach my $p (keys %{$winInstanceRef0})
	{
		my %process = ();
		my $counter0 = $winInstanceRef0->{$p}->{Counters};
		my $counter1 = $winInstanceRef1->{$p}->{Counters};
		foreach my $i (keys %{$counter0}) {
			$process{ COMMAND } = $winInstanceRef0->{$p}->{Name};
			### print "### pingProcessWin: Process running with name $process{COMMAND}\n";
			if ( $process{ COMMAND } eq "_Total" ) {
				%process = ();
				last;
			}
			if($counter0->{$i}->{CounterNameTitleIndex} == $process_data{PID}{ WinCounterIndex } ) {
				$process{ PID } = int($counter0->{$i}->{Counter});
				### print "### pingProcessWin: Found process for $process{ PID }\n";
				$processes{ $process{ PID } } = \%process;
			}
			elsif($counter0->{$i}->{CounterNameTitleIndex} == $process_data{CPU_TIME_MILLIS}{ WinCounterIndex } ){
				my $Numerator0 = $counter0->{$i}->{Counter};
				my $Denominator0 = $winProcRef0->{PerfTime100nSec};
				my $Numerator1 = $counter1->{$i}->{Counter};
				my $Denominator1 = $winProcRef1->{PerfTime100nSec};
				$process{ CPU_PERCENT } = int (($Numerator1 - $Numerator0) / ($Denominator1 - $Denominator0 ) * 100);
				$process{ CPU_TIME_MILLIS } = int($counter1->{$i}->{Counter} / 10000);
			}
			elsif($counter0->{$i}->{CounterNameTitleIndex} == $process_data{MEM_BYTES}{ WinCounterIndex } ) {
				$process{ MEM_BYTES } = int($counter0->{$i}->{Counter});
			}
			elsif($counter0->{$i}->{CounterNameTitleIndex} == $process_data{PAGE_FAULTS_RATE}{ WinCounterIndex } ) {
				my $Numerator0 = $counter0->{$i}->{Counter};
				my $Denominator0 = $winProcRef0->{PerfTime100nSec};
				my $Numerator1 = $counter1->{$i}->{Counter};
				my $Denominator1 = $winProcRef1->{PerfTime100nSec};
				# nano/100 * 10^12 nano/sec = 10^10
				$process{ PAGE_FAULTS_RATE } = int(($Numerator1 - $Numerator0) / ($Denominator1 - $Denominator0 )*10000000000);
			}
		        elsif($counter0->{$i}->{CounterNameTitleIndex} == $process_data{VIRTUAL_MEM_BYTES}{ WinCounterIndex }) {
				$process{ VIRTUAL_MEM_BYTES } = int($counter0->{$i}->{Counter} );
			}
			elsif($counter0->{$i}->{CounterNameTitleIndex} == $process_data{PRI}{ WinCounterIndex }) {
				$process{ PRI } = int($counter0->{$i}->{Counter} );
			}
			elsif($counter0->{$i}->{CounterNameTitleIndex} == $process_data{THREADS}{ WinCounterIndex }) {
				$process{ THREADS } = int($counter0->{$i}->{Counter} );
			}
			elsif($counter0->{$i}->{CounterNameTitleIndex} == $process_data{PRIV_TIME_PERCENT}{ WinCounterIndex }) {
				my $Numerator0 = $counter0->{$i}->{Counter};
				my $Denominator0 = $winProcRef0->{PerfTime100nSec};
				my $Numerator1 = $counter1->{$i}->{Counter};
				my $Denominator1 = $winProcRef1->{PerfTime100nSec};
				$process{ PRIV_TIME_PERCENT } = int (($Numerator1 - $Numerator0) / ($Denominator1 - $Denominator0 ) * 100);
			}
			elsif($counter0->{$i}->{CounterNameTitleIndex} == $process_data{USER_TIME_PERCENT}{ WinCounterIndex }) {
				my $Numerator0 = $counter0->{$i}->{Counter};
				my $Denominator0 = $winProcRef0->{PerfTime100nSec};
				my $Numerator1 = $counter1->{$i}->{Counter};
				my $Denominator1 = $winProcRef1->{PerfTime100nSec};
				$process{ USER_TIME_PERCENT } = int(($Numerator1 - $Numerator0) / ($Denominator1 - $Denominator0 ) * 100);
			}
			elsif($counter0->{$i}->{CounterNameTitleIndex} == $process_data{PAGE_FILE_BYTES}{ WinCounterIndex }) {
				$process{ PAGE_FILE_BYTES } = int($counter0->{$i}->{Counter} );
			}
			elsif($counter0->{$i}->{CounterNameTitleIndex} == $process_data{READ_OPS_RATE}{ WinCounterIndex }) {
				my $Numerator0 = $counter0->{$i}->{Counter};
				my $Denominator0 = $winProcRef0->{PerfTime100nSec};
				my $Numerator1 = $counter1->{$i}->{Counter};
				my $Denominator1 = $winProcRef1->{PerfTime100nSec};
				# nano/100 * 10^12 nano/sec = 10^10
				$process{ READ_OPS_RATE } = int(($Numerator1 - $Numerator0) / ($Denominator1 - $Denominator0 )*10000000000);
			}
			elsif($counter0->{$i}->{CounterNameTitleIndex} == $process_data{WRITE_OPS_RATE}{ WinCounterIndex }) {
				my $Numerator0 = $counter0->{$i}->{Counter};
				my $Denominator0 = $winProcRef0->{PerfTime100nSec};
				my $Numerator1 = $counter1->{$i}->{Counter};
				my $Denominator1 = $winProcRef1->{PerfTime100nSec};
				# nano/100 * 10^12 nano/sec = 10^10
				$process{ WRITE_OPS_RATE } = int(($Numerator1 - $Numerator0) / ($Denominator1 - $Denominator0 )*10000000000);
			}
			elsif($counter0->{$i}->{CounterNameTitleIndex} == $process_data{READ_BYTES_RATE}{ WinCounterIndex }) {
				my $Numerator0 = $counter0->{$i}->{Counter};
				my $Denominator0 = $winProcRef0->{PerfTime100nSec};
				my $Numerator1 = $counter1->{$i}->{Counter};
				my $Denominator1 = $winProcRef1->{PerfTime100nSec};
				# nano/100 * 10^12 nano/sec = 10^10
				$process{ READ_BYTES_RATE } = int(($Numerator1 - $Numerator0) / ($Denominator1 - $Denominator0 )*10000000000);
			}
			elsif($counter0->{$i}->{CounterNameTitleIndex} == $process_data{WRITE_BYTES_RATE}{ WinCounterIndex }) {
				my $Numerator0 = $counter0->{$i}->{Counter};
				my $Denominator0 = $winProcRef0->{PerfTime100nSec};
				my $Numerator1 = $counter1->{$i}->{Counter};
				my $Denominator1 = $winProcRef1->{PerfTime100nSec};
				# nano/100 * 10^12 nano/sec = 10^10
				$process{ WRITE_BYTES_RATE } = int(($Numerator1 - $Numerator0) / ($Denominator1 - $Denominator0 )*10000000000);
				### print "### pingProcessWin: num0=$Numerator0\n";
				### print "### pingProcessWin: num1=$Numerator1\n";
				### print "### pingProcessWin: den0=$Denominator0\n";
				### print "### pingProcessWin: den1=$Denominator1\n";
			}
			elsif($counter0->{$i}->{CounterNameTitleIndex} == $process_data{CONTROL_OPS_RATE}{ WinCounterIndex }) {
				my $Numerator0 = $counter0->{$i}->{Counter};
				my $Denominator0 = $winProcRef0->{PerfTime100nSec};
				my $Numerator1 = $counter1->{$i}->{Counter};
				my $Denominator1 = $winProcRef1->{PerfTime100nSec};
				# nano/100 * 10^12 nano/sec = 10^10
				$process{ CONTROL_OPS_RATE } = int(($Numerator1 - $Numerator0) / ($Denominator1 - $Denominator0 )*10000000000);
			}
		}
		if ( defined $process{PID} ) {
			### print "### pingProcesses: adding PID " . $process{ PID } . "\n";
			$processes{ $process{ PID } } = \%process;
		}
	}
	return %processes;
}

sub defineProcessData {
# Process monitoring data fields, descriptions
# For each hash defined, add to %PROCESS_DATA hash below
#
my %ALIVE = ( description => 'Alive', datatype => 'IntCounter');
my %PID = ( description => 'Process ID', datatype => 'StringEvent', WinCounterIndex => 784);
my %COMMAND = ( description=> 'Startup Command', datatype => 'StringEvent');
my %PPID = ( description=> 'Parent process ID', datatype => 'StringEvent');
my %USER = ( description=> 'Process owner', datatype => 'StringEvent');
my %CPU_PERCENT = ( description=> 'CPU Used vs. available (%)', datatype => 'IntCounter');
my %VIRTUAL_MEM_BYTES = ( description=> 'Virtual memory (bytes)', datatype => 'LongCounter', WinCounterIndex => 186 );
# same as above
#my %PAGE_FILE_BYTES = ( description=> 'Page files (bytes)', datatype => 'LongCounter', WinCounterIndex => 184 );
my %MEM_BYTES = ( description=> 'Resident set size (bytes)', datatype => 'LongCounter', WinCounterIndex => 180 );
my %MEM_PERCENT = ( description=> 'Resident set size (%)', datatype => 'IntCounter');
my %ELAPSED_TIME = ( description=> 'Elapsed Time', datatype => 'StringEvent', WinCounterIndex => 684 );
my %START_TIME = ( description=> 'Start Time', datatype => 'StringEvent'  );
my %BOUND = ( description=> 'Bound to processor (if any)', datatype => 'StringEvent');
my %PRI = ( description=> 'Priority', datatype => 'StringEvent', WinCounterIndex => 682 );
my %STATE = ( description=> 'State', datatype => 'StringEvent');
my %CPU_TIME = ( description=> 'CPU Time', datatype => 'StringEvent' );
my %CPU_TIME_MILLIS = ( description=> 'CPU Time (ms)', datatype => 'LongCounter', WinCounterIndex => 6 );
my %PRIV_TIME_PERCENT = ( description=> 'Privileged Time (%)', datatype => 'IntCounter', WinCounterIndex => 144 );
my %USER_TIME_PERCENT = ( description=> 'User Time (%)', datatype => 'IntCounter', WinCounterIndex => 142 );
my %THREADS = ( description=> 'Kernel threads', datatype => 'IntCounter', WinCounterIndex => 680 );
my %THREADS_SUSPENDED = ( description=> 'Kernel threads suspended', datatype => 'IntCounter');
my %PAGE_FAULTS_RATE = ( description=> 'Page Faults per sec', datatype => 'IntCounter', WinCounterIndex => 28 );
my %CONTROL_OPS_RATE = ( description=> 'Control ops per sec', datatype => 'IntCounter', WinCounterIndex => 1414 );
my %WRITE_BYTES_RATE = ( description=> 'Write bytes per sec', datatype => 'LongCounter', WinCounterIndex => 1418 );
my %READ_BYTES_RATE = ( description=> 'Read bytes per sec', datatype => 'LongCounter', WinCounterIndex => 1416 );
my %WRITE_OPS_RATE = ( description=> 'Write ops per sec', datatype => 'IntCounter', WinCounterIndex => 1412 );
my %READ_OPS_RATE = ( description=> 'Read ops per sec', datatype => 'IntCounter', WinCounterIndex => 1410 );

my %PROCESS_DATA = (
	ALIVE => \%ALIVE,
	PID => \%PID,
	COMMAND => \%COMMAND,
	PPID => \%PPID,
	USER => \%USER,
	CPU_PERCENT => \%CPU_PERCENT,
	VIRTUAL_MEM_BYTES => \%VIRTUAL_MEM_BYTES,
#	PAGE_FILE_BYTES => \%PAGE_FILE_BYTES,
	MEM_BYTES => \%MEM_BYTES,
	MEM_PERCENT => \%MEM_PERCENT,
	ELAPSED_TIME => \%ELAPSED_TIME,
	START_TIME => \%START_TIME,
	BOUND => \%BOUND,
	PRI => \%PRI,
	STATE => \%STATE,
	CPU_TIME => \%CPU_TIME,
	CPU_TIME_MILLIS => \%CPU_TIME_MILLIS,
	PRIV_TIME_PERCENT => \%PRIV_TIME_PERCENT,
	USER_TIME_PERCENT => \%USER_TIME_PERCENT,
	THREADS => \%THREADS,
	THREADS_SUSPENDED => \%THREADS_SUSPENDED,
	PAGE_FAULTS_RATE => \%PAGE_FAULTS_RATE,
	CONTROL_OPS_RATE=> \%CONTROL_OPS_RATE,
	WRITE_BYTES_RATE=> \%WRITE_BYTES_RATE,
	READ_BYTES_RATE=> \%READ_BYTES_RATE,
	WRITE_OPS_RATE=> \%WRITE_OPS_RATE,
	READ_OPS_RATE=> \%READ_OPS_RATE,
);
return %PROCESS_DATA;
}

1;